# -*- coding: utf8 -*-

import tensorflow as tf
import numpy as np

def flatten(x):
    return tf.reshape(x, [-1, np.prod(x.get_shape().as_list()[1:])])

def normalized_columns_initializer(std: float = 1.0):
    def _initializer(shape, dtype=None, partition_info=None):
        out = np.random.randn(*shape).astype(np.float32)
        out *= std / np.sqrt(np.square(out).sum(axis=0, keepdims=True))
        return tf.constant(out)
    return _initializer

def fan_in_initializer(fan_in_size: int):
    def _initializer(shape, dtype=None, partition_info=None):
        bound = 1 / np.sqrt(fan_in_size)
        out = np.random.uniform(-bound, bound, shape)
        return tf.constant(out.astype(np.float32))
    return _initializer


def linear_fan_in(x, output_size: int):
    input_size = x.shape[1].value
    w = tf.get_variable("w", [input_size, output_size], initializer=fan_in_initializer(input_size))
    b = tf.get_variable("b", [output_size], initializer=fan_in_initializer(input_size))
    x = tf.nn.xw_plus_b(x, w, b)
    return x, [w, b]


def linear(x, size, name, initializer=None, bias_init=None):
    if bias_init is None:
        bias_init = tf.constant_initializer(0.0)
    w = tf.get_variable(name + "/w", [x.get_shape()[1], size], initializer=initializer)
    b = tf.get_variable(name + "/b", [size], initializer=bias_init)
    return tf.matmul(x, w) + b

def conv2d(x, num_filters, name, filter_size=(3, 3), stride=(1, 1), pad="SAME", dtype=tf.float32, collections=None):
    """
    2-dimensional convolutional layer.
    Source: https://github.com/openai/universe-starter-agent/blob/a3fdfba297c8c24d62d3c53978fb6fb26f80e76e/model.py
    """

    with tf.variable_scope(name):
        stride_shape = [1, stride[0], stride[1], 1]
        filter_shape = [filter_size[0], filter_size[1], int(x.get_shape()[3]), num_filters]

        # there are "num input feature maps * filter height * filter width"
        # inputs to each hidden unit
        fan_in = np.prod(filter_shape[:3])
        # each unit in the lower layer receives a gradient from:
        # "num output feature maps * filter height * filter width" /
        #   pooling size
        fan_out = np.prod(filter_shape[:2]) * num_filters
        # initialize weights with random weights
        w_bound = np.sqrt(6. / (fan_in + fan_out))

        w = tf.get_variable("W", filter_shape, dtype, tf.random_uniform_initializer(-w_bound, w_bound),
                            collections=collections)
        b = tf.get_variable("b", [1, 1, 1, num_filters], initializer=tf.constant_initializer(0.0),
                            collections=collections)
        return tf.nn.conv2d(x, w, stride_shape, pad) + b

def mu_sigma_layer(inputs, n_outputs: int):
    """
    Create a layer that makes a mu and sigma,
    e.g. to use in continuous action spaces.
    """

    mu = tf.contrib.layers.fully_connected(
        inputs=inputs,
        num_outputs=n_outputs,
        activation_fn=None,
        weights_initializer=normalized_columns_initializer(0.01),
        biases_initializer=tf.zeros_initializer(),
        scope="mu")
    mu = tf.squeeze(mu, name="mu")

    sigma = tf.contrib.layers.fully_connected(
        inputs=inputs,
        num_outputs=n_outputs,
        activation_fn=None,
        weights_initializer=normalized_columns_initializer(0.01),
        biases_initializer=tf.zeros_initializer(),
        scope="sigma")
    sigma = tf.squeeze(sigma)
    sigma = tf.nn.softplus(sigma) + 1e-5
    return mu, sigma

def create_accumulative_gradients_op(net_vars, identifier: int = 0):
    """Make an operation to create accumulative gradients"""
    accum_grads = []
    with tf.name_scope(name="create_accum_{}".format(identifier), values=net_vars):
        for var in net_vars:
            zero = tf.zeros(var.get_shape().as_list(), dtype=var.dtype)
            name = var.name.replace(":", "_") + "_accum_grad"
            accum_grad = tf.Variable(zero, name=name, trainable=False)
            accum_grads.append(accum_grad)
        return accum_grads

def add_accumulative_gradients_op(net_vars, accum_grads, loss, identifier: int = 0):
    """Make an operation to add a gradient to the total"""
    accum_grad_ops = []
    with tf.name_scope(name="grad_ops_{}".format(identifier), values=net_vars):
        grads = tf.gradients(loss, net_vars, gate_gradients=False,
                             aggregation_method=None,
                             colocate_gradients_with_ops=False)
    with tf.name_scope(name="accum_ops_{}".format(identifier), values=[]):
        for (grad, var, accum_grad) in zip(grads, net_vars, accum_grads):
            name = var.name.replace(":", "_") + "_accum_grad_ops"
            accum_ops = tf.assign_add(accum_grad, grad, name=name)
            accum_grad_ops.append(accum_ops)
        return tf.group(*accum_grad_ops, name="accum_group_{}".format(identifier))

def reset_accumulative_gradients_op(net_vars, accum_grads, identifier: int = 0):
    """Make an operation to reset the accumulation to zero."""
    reset_grad_ops = []
    with tf.name_scope(name="reset_grad_ops_{}".format(identifier), values=net_vars):
        for (var, accum_grad) in zip(net_vars, accum_grads):
            zero = tf.zeros(var.get_shape().as_list(), dtype=var.dtype)
            name = var.name.replace(":", "_") + "_reset_grad_ops"
            reset_ops = tf.assign(accum_grad, zero, name=name)
            reset_grad_ops.append(reset_ops)
        return tf.group(*reset_grad_ops, name="reset_accum_group_{}".format(identifier))

def create_sync_net_op(source_vars, target_vars):
    return tf.group(*[v1.assign(v2) for v1, v2 in zip(target_vars, source_vars)], name="sync_net")

def batch_norm_layer(x, training_phase, scope_bn: str, activation=None):
    return tf.cond(
        training_phase,
        lambda: tf.contrib.layers.batch_norm(
            x,
            activation_fn=activation,
            center=True,
            scale=True,
            updates_collections=None,
            is_training=True,
            reuse=None,
            scope=scope_bn,
            decay=0.9,
            epsilon=1e-5),
        lambda: tf.contrib.layers.batch_norm(
            x,
            activation_fn=activation,
            center=True,
            scale=True,
            updates_collections=None,
            is_training=False,
            reuse=True,
            scope=scope_bn,
            decay=0.9,
            epsilon=1e-5))


def kl_divergence(logits1, logits2):
    a0 = logits1 - tf.reduce_max(logits1, axis=-1, keepdims=True)
    a1 = logits2 - tf.reduce_max(logits2, axis=-1, keepdims=True)
    ea0 = tf.exp(a0)
    ea1 = tf.exp(a1)
    z0 = tf.reduce_sum(ea0, axis=-1, keepdims=True)
    z1 = tf.reduce_sum(ea1, axis=-1, keepdims=True)
    p0 = ea0 / z0
    return tf.reduce_sum(p0 * (a0 - tf.log(z0) - a1 + tf.log(z1)), axis=-1)



# -*- coding: utf8 -*-

# Inspired by https://github.com/openai/gym/blob/master/gym/envs/registration.py
# When making an environment, we first look if we registered a version of it ourselves.
# Else, we make just make one using the Environment class.
from typing import Sequence
import numpy as np

import gym
from yarll.misc.utils import AtariRescale42x42
from yarll.environment.environment import Environment

gym.logger.set_level(gym.logger.ERROR)

class EnvSpec(gym.envs.registration.EnvSpec):
    """
    Modified version of `gym.envs.registration.EnvSpec`
    to allow env initalization with different parameters.
    """

    def make(self, **kwargs):
        """Instantiates an instance of the environment with appropriate kwargs"""
        if self._entry_point is None:
            raise gym.error.Error(
                "Attempting to make deprecated env {}. (HINT: is there a newer registered version of this env?)".format(self.id))

        elif callable(self._entry_point):
            env = self._entry_point()
        else:
            all_kwargs = self._kwargs
            all_kwargs.update(kwargs)
            cls = gym.envs.registration.load(self._entry_point)
            if cls == Environment:
                env = cls(gym.make(all_kwargs["old_env_name"]), **all_kwargs)
            else:
                env = cls(**all_kwargs)

        # Make the enviroment aware of which spec it came from.
        env.unwrapped.spec = self

        return env

def make(env_id: str, **kwargs):
    spec = gym.envs.registry.spec(env_id)
    env = spec.make(**kwargs)

    if not isinstance(env, Environment):
        if (env.spec.timestep_limit is not None) and not spec.tags.get('vnc'):
            from gym.wrappers.time_limit import TimeLimit
            env = TimeLimit(env,
                            max_episode_steps=env.spec.max_episode_steps,
                            max_episode_seconds=env.spec.max_episode_seconds)
        env = Environment(env)
    if "atari.atari_env" in env.unwrapped.__module__:
        env = AtariRescale42x42(env)
    if "wrapper_entry_points" in spec.tags:
        for wrapper_info in spec.tags["wrapper_entry_points"]:
            kwargs = {}
            if isinstance(wrapper_info, str):
                cls = gym.envs.registration.load(wrapper_info)
            else:
                cls = gym.envs.registration.load(wrapper_info["entry_point"])
                kwargs = wrapper_info["kwargs"]
            env = cls(env, **kwargs)
    return env

def make_environments(descriptions: Sequence[dict]) -> list:
    """Make environments using a list of descriptions."""
    return [make(**d) for d in descriptions]

def make_random_environments(env_id: str, n_envs: int) -> list:
    """Make n_envs random environments of the env_name class."""
    spec = gym.envs.registry.spec(env_id)
    cls = gym.envs.registration.load(spec._entry_point)
    envs = []
    for _ in range(n_envs):
        args = {"id": env_id}
        for p in cls.changeable_parameters:
            if p["type"] == "range":  # Assume for now only range parameters are used
                args[p["name"]] = np.random.uniform(p["low"], p["high"])
            else:
                raise NotImplementedError("Only able to make environments with range parameters.")
        envs.append(make(**args))
    for env in envs:
        print(id(env.metadata))
    return envs



# -*- coding: utf8 -*-

import numpy as np
from yarll.actionselection.action_selection import ActionSelection

class CategoricalActionSelection(ActionSelection):
    pass

class ProbabilisticCategoricalActionSelection(CategoricalActionSelection):
    """Sample from categorical distribution, specified by a vector of class probabilities"""

    @staticmethod
    def select_action(probabilities):
        return np.random.choice(len(probabilities), p=probabilities)

class MaxCategoricalActionSelection(CategoricalActionSelection):
    """Choose the action with the highest probability."""

    @staticmethod
    def select_action(probabilities):
        return np.argmax(probabilities)



#!/usr/bin/env python
# -*- coding: utf8 -*-

import argparse
import json

predef = {
    "experiment_name": "CartPole-KT",
    "environments": {
        "type": "json",
        "source": "Environment/random_CartPole_envs4.json"
    },
    "agent": {
        "name": "AsyncKnowledgeTransfer",
        "args": {
            "monitor_path": "/tmp/CartPole-v0-AKT",
            "video": False,
            "save_model": True,
            "monitor": True,
            "n_iter": 200,
            "switch_at_iter": 100,
            "learning_rate": 0.05
        }
    }
}

parser = argparse.ArgumentParser()
parser.add_argument("agent", choices=["KnowledgeTransfer", "AsyncKnowledgeTransfer"])
parser.add_argument("envs", type=str)
parser.add_argument("destination", type=str)

def main():
    args = parser.parse_args()
    predef["agent"]["name"] = args.agent
    predef["environments"]["source"] = args.envs
    with open(args.destination, "w") as f:
        json.dump(predef, f)

if __name__ == '__main__':
    main()



# -*- coding: utf8 -*-

from yarll.policies.policy import Policy

import numpy as np

class EGreedy(Policy):
    """Take the best action with a probability and a random one otherwise."""
    def __init__(self, epsilon):
        super(EGreedy, self).__init__()
        self.epsilon = epsilon

    def select_action(self, values):
        idx = None
        if (np.random.rand() < self.epsilon):  # With a probability of epsilon...
            idx = np.random.randint(len(values))  # Choose a random action...
        else:
            idx = np.argmax(values)  # Else return the action with the highest value
        return idx, values[idx]  # Return the action and the associated value



# -*- coding: utf8 -*-

#  Cross-Entropy Method
#  Source: http://rl-gym-doc.s3-website-us-west-2.amazonaws.com/mlss/lab1.html

import numpy as np
from gym import wrappers
from gym.spaces import Discrete, Box

from yarll.agents.agent import Agent
from yarll.misc.exceptions import WrongShapeError

# ================================================================
# Policies
# ================================================================

class Policy(object):
    """
    Cross-Entropy Method policy.
    """

    def act(self, ob):
        """
        Decide which action to take given an observation.
        """
        raise NotImplementedError()

class DeterministicDiscreteActionLinearPolicy(Policy):
    """Deterministicially select an action from a discrete action space using a linear function."""
    def __init__(self, theta, ob_space, ac_space) -> None:
        """
        dim_ob: dimension of observations
        n_actions: number of actions
        theta: flat vector of parameters
        """
        dim_ob = ob_space.shape[0]
        n_actions = ac_space.n
        expected_shape = (dim_ob + 1) * n_actions
        if len(theta) != expected_shape:
            raise WrongShapeError("Expected a theta of length {} instead of {}".format(expected_shape, len(theta)))
        self.W = theta[0: dim_ob * n_actions].reshape(dim_ob, n_actions)
        self.b = theta[dim_ob * n_actions: None].reshape(1, n_actions)

    def act(self, ob):
        """Select the action that got the highest value from the linear function."""
        y = ob.dot(self.W) + self.b
        a = y.argmax()
        return a

class DeterministicContinuousActionLinearPolicy(Policy):
    """Deterministicially select an action from a continuous action space using a linear function."""
    def __init__(self, theta, ob_space, ac_space) -> None:
        """
        dim_ob: dimension of observations
        dim_ac: dimension of action vector
        theta: flat vector of parameters
        """
        self.ac_space = ac_space
        dim_ob = ob_space.shape[0]
        dim_ac = ac_space.shape[0]
        expected_shape = (dim_ob + 1) * dim_ac
        if len(theta) != expected_shape:
            raise WrongShapeError("Expected a theta of length {} instead of {}".format(expected_shape, len(theta)))
        self.W = theta[0:dim_ob * dim_ac].reshape(dim_ob, dim_ac)
        self.b = theta[dim_ob * dim_ac:None]

    def act(self, ob):
        a = np.clip(ob.dot(self.W) + self.b, self.ac_space.low, self.ac_space.high)
        return a

class CEM(Agent):
    """Cross-Entropy Method learner"""
    def __init__(self, env, monitor_path: str, video: bool = True, **usercfg) -> None:
        super(CEM, self).__init__(**usercfg)
        self.env = wrappers.Monitor(env, monitor_path, force=True, video_callable=(None if video else False))
        self.config.update(dict(
            num_steps=env.spec.tags.get("wrapper_config.TimeLimit.max_episode_steps"),  # maximum length of episode
            n_iter=100,  # number of iterations of CEM
            batch_size=25,  # number of samples per batch
            elite_frac=0.2  # fraction of samples used as elite set
        ))
        self.config.update(usercfg)
        if isinstance(env.action_space, Discrete):
            self.dim_theta = (env.observation_space.shape[0] + 1) * env.action_space.n
        elif isinstance(env.action_space, Box):
            self.dim_theta = (env.observation_space.shape[0] + 1) * env.action_space.shape[0]
        else:
            raise NotImplementedError
        # Initialize mean and standard deviation
        self.theta_mean = np.zeros(self.dim_theta)
        self.theta_std = np.ones(self.dim_theta)

    def make_policy(self, theta) -> Policy:
        if isinstance(self.env.action_space, Discrete):
            return DeterministicDiscreteActionLinearPolicy(theta, self.env.observation_space, self.env.action_space)
        elif isinstance(self.env.action_space, Box):
            return DeterministicContinuousActionLinearPolicy(theta, self.env.observation_space, self.env.action_space)
        else:
            raise NotImplementedError

    def noisy_evaluation(self, theta) -> float:
        policy: Policy = self.make_policy(theta)
        rew = self.do_episode(policy)
        return rew

    def do_episode(self, policy: Policy, render=False):
        total_rew = 0
        ob = self.env.reset()
        for _ in range(self.config["num_steps"]):
            a = policy.act(ob)
            ob, reward, done, _info = self.env.step(a)
            if render:
                self.env.render()
            total_rew += reward
            if done:
                break
        return total_rew

    def learn(self):
        for iteration in range(self.config["n_iter"]):
            # Sample parameter vectors
            thetas = [np.random.normal(self.theta_mean, self.theta_std, self.dim_theta)
                      for _ in range(self.config["batch_size"])]
            rewards = [self.noisy_evaluation(theta) for theta in thetas]
            # Get elite parameters
            n_elite = int(self.config["batch_size"] * self.config["elite_frac"])
            elite_inds = np.argsort(rewards)[self.config["batch_size"] - n_elite:self.config["batch_size"]]
            elite_thetas = [thetas[i] for i in elite_inds]
            # Update theta_mean, theta_std
            self.theta_mean = np.mean(elite_thetas, axis=0)
            self.theta_std = np.std(elite_thetas, axis=0)
            print("iteration {:d}. mean f: {:>8.3g}. max f: {:>8.3g}".format(
                iteration,
                np.mean(rewards),
                np.max(rewards)))
            self.do_episode(self.make_policy(self.theta_mean))
        self.do_episode(self.make_policy(self.theta_mean), render=True)



# -*- coding: utf8 -*-

from yarll.environment.wrappers import DescriptionWrapper

class Environment(DescriptionWrapper):
    def __init__(self, env, **kwargs):
        super(Environment, self).__init__(env)



#!/usr/bin/env python
# -*- coding: utf8 -*-

import argparse
import os
import time
import tensorflow as tf

tf.logging.set_verbosity(tf.logging.ERROR)
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '5'

from yarll.misc.utils import cluster_spec

parser = argparse.ArgumentParser()
parser.add_argument("n_tasks", type=int, help="Total number of tasks in this experiment.")
parser.add_argument("--n_masters", type=int, default=0, help="Number of masters")
parser.add_argument("--task_id", type=int, default=0, help="ID of this task.")

def main(_):
    args = parser.parse_args()
    spec = cluster_spec(args.n_tasks, 1, args.n_masters)
    cluster = tf.train.ClusterSpec(spec).as_cluster_def()
    server = tf.train.Server(  # pylint: disable=W0612
        cluster,
        job_name="ps",
        task_index=args.task_id,
        config=tf.ConfigProto(device_filters=["/job:ps"]))
    while True:
        time.sleep(1000)


if __name__ == '__main__':
    tf.app.run()



# -*- coding: utf8 -*-

from yarll.actionselection.action_selection import ActionSelection
import numpy as np

class ContinuousActionSelection(ActionSelection):
    """Selection of an action in a continuous action space"""

    # def select_action(self, mu, sigma):
    #     return np.random.normal(mu, sigma)

    # Without sigma
    def select_action(self, mu):
        return np.random.normal(mu)



#!/usr/bin/env python
# -*- coding: utf8 -*-

import os
import argparse
import json
import matplotlib.pyplot as plt

parser = argparse.ArgumentParser()
parser.add_argument("envs_dir", type=str, help="Directory with environment specification files")

def main():
    args = parser.parse_args()

    _, _, env_spec_files = next(os.walk(args.envs_dir))
    values = None
    for env_spec_file in env_spec_files:
        with open(os.path.join(args.envs_dir, env_spec_file)) as f:
            envs = json.load(f)
            if values is None:
                values = [{} for _ in range(len(envs))]
            for i, env in enumerate(envs):
                for key, value in env.items():
                    if key == "name":
                        continue
                    values[i].setdefault(key, []).append(value)
    keys = values[0].keys()
    for key in keys:
        fig = plt.figure()
        plt.boxplot([env[key] for env in values])
        plt.title(key)
        fig.canvas.set_window_title(key)
    plt.show()

if __name__ == '__main__':
    main()



# -*- coding: utf8 -*-

from yarll.agents.ppo.dppo import DPPO
from yarll.agents.trpo.trpo import trpo_loss
from yarll.agents.actorcritic.actor_critic import ActorCriticNetworkDiscrete,\
    ActorCriticNetworkDiscreteCNN, ActorCriticNetworkContinuous

class DTRPO(DPPO):
    """Trust Region Policy Optimization agent."""

    def __init__(self, env, monitor_path: str, **usercfg) -> None:
        usercfg["kl_coef"] = 1.0  # beta
        super(DTRPO, self).__init__(env, monitor_path, **usercfg)

    def make_actor_loss(self, old_network, new_network, advantage):
        return trpo_loss(old_network.action_log_prob, new_network.action_log_prob, self.config["kl_coef"], advantage)


class DTRPODiscrete(DTRPO):
    def build_networks(self) -> ActorCriticNetworkDiscrete:
        return ActorCriticNetworkDiscrete(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]),
            int(self.config["n_hidden_layers"]))


class DTRPODiscreteCNN(DTRPODiscrete):
    def build_networks(self) -> ActorCriticNetworkDiscreteCNN:
        return ActorCriticNetworkDiscreteCNN(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]))


class DTRPOContinuous(DTRPO):
    def build_networks(self) -> ActorCriticNetworkContinuous:
        return ActorCriticNetworkContinuous(
            list(self.env.observation_space.shape),
            self.env.action_space,
            int(self.config["n_hidden_units"]),
            int(self.config["n_hidden_layers"]))

    def get_env_action(self, action):
        return action



#!/usr/bin/env python
# -*- coding: utf8 -*-

import os
import queue
import argparse
import time
import tensorflow as tf
import numpy as np

tf.logging.set_verbosity(tf.logging.ERROR)
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '5'

from yarll.environment.registration import make
from yarll.misc.network_ops import factorized_conv2d, flatten, normalized_columns_initializer, create_sync_net_op, add_initializers
from yarll.misc.utils import discount_rewards, FastSaver, load, json_to_dict, cluster_spec
from yarll.agents.actorcritic.actor_critic import actor_critic_discrete_loss
from yarll.agents.actorcritic.a3c_worker import RunnerThread


class AKTTask(object):
    """Asynchronous knowledge transfer learner thread. Used to learn using one specific variation of a task."""

    def __init__(self, env, task_id, target_task, cluster, monitor_path, config, video=False):
        super(AKTTask, self).__init__()
        self.env = env
        self.task_id = task_id
        self.target_task = target_task
        self.monitor_path = monitor_path
        self.config = config
        self.add_accum_grad = None  # To be filled in later

        # Only used (and overwritten) by agents that use an RNN
        self.initial_features = None

        self.first_source_or_target_task = self.task_id == 0 or (
            self.target_task and self.task_id == self.config["n_tasks"] - self.config["n_target_tasks"])

        state_shape = env.observation_space.shape
        self.states = tf.placeholder(tf.float32, [None] + list(state_shape), name="states")
        self.adv = tf.placeholder(tf.float32, name="advantage")
        self.actions_taken = tf.placeholder(tf.float32, name="actions_taken")
        self.r = tf.placeholder(tf.float32, [None], name="r")
        worker_device = "/job:worker/task:{}/cpu:0".format(task_id)
        # Global network
        shared_device = tf.train.replica_device_setter(1, worker_device=worker_device)
        with tf.device(shared_device):
            with tf.variable_scope("global"):
                self.build_shared_network()
                self._global_step = tf.get_variable(
                    "global_step",
                    [],
                    tf.int32,
                    initializer=tf.constant_initializer(0, dtype=tf.int32),
                    trainable=False)
                self.global_shared_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES,
                                                            tf.get_variable_scope().name)

        with tf.device(worker_device):
            with tf.variable_scope("local"):
                with tf.variable_scope("shared"):
                    # Part of the network that is shared between all tasks (by copying from global)
                    knowledge_bases = self.build_shared_network()
                    self.local_shared_vars = tf.get_collection(
                        tf.GraphKeys.TRAINABLE_VARIABLES, tf.get_variable_scope().name)
                with tf.variable_scope("task"):
                    # Part of the network that is specific for the task and not shared
                    self.action, self.value, entropy, self.loss, \
                    self.losses, self.sparse_representations = self.build_task_network(knowledge_bases)
                    # Add L1 loss for sparse representations weights of source tasks
                    if not self.target_task and self.config["l1_coef"] > 0.0:
                        regularizer = tf.contrib.layers.l1_regularizer(self.config["l1_coef"])
                        self.l1_loss = 0
                        for w in self.sparse_representations:
                            self.l1_loss += regularizer(w)
                        self.loss += self.l1_loss
                        self.losses[-1] = self.loss
                    self.task_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, tf.get_variable_scope().name)

                with tf.variable_scope("optimizer"):
                    optimizer = tf.train.AdamOptimizer(self.config["learning_rate"], name="optim")
                    grad_vars = (self.local_shared_vars if not(target_task) else []) + self.task_vars
                    grads = tf.gradients(self.loss, grad_vars)
                    grads, _ = tf.clip_by_global_norm(grads, self.config["gradient_clip_value"])

                    # Apply gradients to the weights of the master network
                    apply_vars = (self.global_shared_vars if not(target_task) else []) + self.task_vars
                    apply_grads = optimizer.apply_gradients(zip(grads, apply_vars))
                    self.optimizer_vars = tf.get_collection(
                        tf.GraphKeys.GLOBAL_VARIABLES, tf.get_variable_scope().name)

            self.n_steps = tf.shape(self.states)[0]
            inc_step = self._global_step.assign_add(self.n_steps)

            self.train_op = tf.group(apply_grads, inc_step, name="train")

            self.sync_net = create_sync_net_op(self.global_shared_vars, self.local_shared_vars)

            n_steps = tf.to_float(self.n_steps)
            actor_loss_summary = tf.summary.scalar("model/actor_loss", self.losses[0] / n_steps)
            critic_loss_summary = tf.summary.scalar("model/critic_loss", self.losses[1] / n_steps)
            id_loss_summary = tf.summary.scalar("model/id_loss", self.losses[2])
            entropy_summary = tf.summary.scalar("model/entropy", -tf.reduce_mean(entropy))
            mean, std = tf.nn.moments(self.value, axes=[0])
            mean_q_summary = tf.summary.scalar("model/q/mean", mean)
            std_q_summary = tf.summary.scalar("model/q/std", std)
            loss_summary = tf.summary.scalar("model/loss", self.loss / n_steps)
            summary_grad_norm = tf.summary.scalar("model/grad_global_norm", tf.global_norm(grads))
            summary_var_norm = tf.summary.scalar("model/var_global_norm",
                                                 tf.global_norm(self.local_shared_vars + self.task_vars))
            summaries = [actor_loss_summary, critic_loss_summary, id_loss_summary, loss_summary,
                         entropy_summary,
                         mean_q_summary, std_q_summary,
                         summary_grad_norm, summary_var_norm]
            if not self.target_task and self.config["l1_coef"] > 0.0:
                l1_loss_summary = tf.summary.scalar("model/l1_loss", self.l1_loss)
                summaries.append(l1_loss_summary)
            self.summary_op = tf.summary.merge(summaries)

            if self.first_source_or_target_task:
                histogram_summaries = []
                for v in self.global_shared_vars:
                    histogram_summaries.append(tf.summary.histogram(v.name, v))
                for v in self.task_vars:
                    histogram_summaries.append(tf.summary.histogram("task{}/{}".format(self.task_id, v.name), v))
                self.histogram_summary_op = tf.summary.merge(histogram_summaries)

        self.runner = RunnerThread(self.env, self, self.config["n_local_steps"], task_id == 0 and video)

        variables_to_save = self.global_shared_vars + self.task_vars
        init_op = tf.variables_initializer(self.global_shared_vars + [self._global_step])
        init_all_op = tf.global_variables_initializer()
        saver = FastSaver(variables_to_save)

        # Write the summary of each task in a different directory
        self.writer = tf.summary.FileWriter(os.path.join(monitor_path, "task{}".format(self.task_id)))

        self.server = tf.train.Server(
            cluster,
            job_name="worker",
            task_index=task_id,
            config=tf.ConfigProto(intra_op_parallelism_threads=1, inter_op_parallelism_threads=2)
        )

        local_init_op = tf.variables_initializer(self.task_vars + self.optimizer_vars)

        def init_fn(scaffold, sess):
            sess.run(init_all_op)

        self.scaffold = tf.train.Scaffold(
            init_op=init_op if not(self.target_task) else self.sync_net,
            ready_for_local_init_op=tf.report_uninitialized_variables(self.global_shared_vars),
            local_init_op=local_init_op,
            init_fn=init_fn if not(self.target_task) else None,
            saver=saver,
            ready_op=tf.report_uninitialized_variables(variables_to_save)
        )

        self.config_proto = tf.ConfigProto(
            device_filters=["/job:ps", "/job:worker/task:{}/cpu:0".format(task_id)])

        self.session = None
        self.local_steps = 0 # Number of updates done

    def build_shared_network(self):
        raise NotImplementedError()

    def build_task_network(self, knowledge_bases):
        raise NotImplementedError()

    @property
    def global_step(self):
        return self._global_step.eval(session=self.session)

    def get_critic_value(self, states, features):
        feed_dict = {
            self.states: states
        }
        if features is not None:
            feed_dict[self.rnn_state_in] = features
        return self.session.run(self.value, feed_dict=feed_dict)[0]

    def choose_action(self, state, features):
        """Choose an action."""
        feed_dict = {
            self.states: [state]
        }
        fetches = [self.action, self.value]
        if features is not None:
            feed_dict[self.rnn_state_in] = features
            fetches += [self.rnn_state_out]
        outputs = self.session.run(fetches, feed_dict=feed_dict)
        return dict(zip(["action", "value", "features"], outputs))

    def get_env_action(self, action):
        return np.argmax(action)

    def pull_batch_from_queue(self):
        """
        Take a trajectory from the queue.
        Also immediately try to extend it if the episode
        wasn't over and more transitions are available.
        """
        trajectory = self.runner.queue.get(timeout=600.0)
        while not trajectory.terminal:
            try:  # Maybe the full trajectory is already available
                trajectory.extend(self.runner.queue.get_nowait())
            except queue.Empty:
                break
        return trajectory

    def learn(self):
        # Assume global shared parameter vectors θ and θv and global shared counter T = 0
        # Assume thread-specific parameter vectors θ' and θ'v
        if self.task_id != 0 and not self.target_task:
            time.sleep(5)
        stop_at = self.config["switch_at_steps"] if not(self.target_task) else self.config["T_max"]
        with tf.train.MonitoredTrainingSession(
            master=self.server.target,
            is_chief=self.first_source_or_target_task,
            config=self.config_proto,
            save_summaries_secs=30,
            scaffold=self.scaffold
        ) as sess:
            tf.get_default_graph().finalize()
            self.session = sess
            sess.run(self.sync_net) # Runner may use vars before the sync net in the while loop
            self.runner.start_runner(sess, self.writer)
            while not sess.should_stop() and self.global_step < stop_at:
                # Synchronize thread-specific parameters θ' = θ and θ'v = θv
                sess.run(self.sync_net)
                trajectory = self.pull_batch_from_queue()
                v = 0 if trajectory.terminal else self.get_critic_value(
                    np.asarray(trajectory.states)[None, -1], trajectory.features[-1])
                rewards_plus_v = np.asarray(trajectory.rewards + [v])
                vpred_t = np.asarray(trajectory.values + [v])
                delta_t = trajectory.rewards + self.config["gamma"] * vpred_t[1:] - vpred_t[:-1]
                batch_r = discount_rewards(rewards_plus_v, self.config["gamma"])[:-1]
                batch_adv = discount_rewards(delta_t, self.config["gamma"])
                fetches = [self.train_op]
                should_make_summaries = self.local_steps % 11 == 0
                if should_make_summaries:
                    if self.first_source_or_target_task:
                        fetches.append(self.histogram_summary_op)
                    fetches.append(self.summary_op)
                    fetches.append(self._global_step)
                states = np.asarray(trajectory.states)
                feed_dict = {
                    self.states: states,
                    self.actions_taken: np.asarray(trajectory.actions),
                    self.adv: batch_adv,
                    self.r: np.asarray(batch_r)
                }
                feature = trajectory.features[0]
                if feature != [] and feature is not None:
                    feed_dict[self.rnn_state_in] = feature
                results = sess.run(fetches, feed_dict)
                if should_make_summaries:
                    global_step = results[-1]
                    self.writer.add_summary(results[-2], global_step)
                    if self.first_source_or_target_task:
                        self.writer.add_summary(results[-3], global_step)
                    self.writer.flush()
                self.local_steps += 1
            self.runner.stop_requested = True

class AKTTaskDiscrete(AKTTask):
    def __init__(self, env, task_id, target_task, cluster, monitor_path, config, video=False):
        self.nA = env.action_space.n
        super(AKTTaskDiscrete, self).__init__(env, task_id, target_task, cluster, monitor_path, config, video=video)

    def build_shared_network(self):
        knowledge_base = tf.get_variable("knowledge_base",
                                         [self.states.get_shape()[-1].value, self.config["n_hidden_units"]],
                                         initializer=normalized_columns_initializer(0.01))
        return [knowledge_base]

    def build_task_network(self, knowledge_bases):
        knowledge_base = knowledge_bases[0]
        sparse_representation_L1 = tf.get_variable(
            "L1/sparse",
            [self.config["n_hidden_units"], self.config["n_hidden_units"]],
            initializer=add_initializers(normalized_columns_initializer(0.01),
                                         tf.constant_initializer(np.identity(self.config["n_hidden_units"],
                                                                             dtype=np.float32))))
            # initializer=normalized_columns_initializer(0.01))
        L1_bias = tf.get_variable(
            "L1/b",
            [self.config["n_hidden_units"]],
            initializer=tf.constant_initializer(0.0))

        W = tf.matmul(knowledge_base, sparse_representation_L1)
        L1 = tf.tanh(tf.nn.xw_plus_b(self.states, W, L1_bias))

        logits = tf.contrib.layers.fully_connected(
            inputs=L1,
            num_outputs=self.nA,
            activation_fn=None,
            weights_initializer=normalized_columns_initializer(0.01),
            biases_initializer=tf.zeros_initializer(),
            scope="logits"
        )

        value = tf.contrib.layers.fully_connected(
            inputs=L1,
            num_outputs=1,
            activation_fn=None,
            weights_initializer=normalized_columns_initializer(1.0),
            biases_initializer=tf.zeros_initializer(),
            scope="value"
        )
        value = tf.reshape(value, [-1], name="value")
        sparse_representations = [sparse_representation_L1]

        probs = tf.nn.softmax(logits, name="probs")

        action_entropy = probs * tf.nn.log_softmax(logits)

        action = tf.squeeze(tf.multinomial(logits - tf.reduce_max(logits, [1], keepdims=True), 1), [1])
        action = tf.squeeze(tf.one_hot(action, self.nA), name="action")

        actor_loss, critic_loss, loss = actor_critic_discrete_loss(logits,
                                                                   probs,
                                                                   value,
                                                                   self.actions_taken,
                                                                   self.adv,
                                                                   self.r,
                                                                   vf_coef=self.config["vf_coef"])
        id_loss = tf.reduce_sum(tf.square(sparse_representation_L1 - tf.constant(np.identity(self.config["n_hidden_units"], dtype=np.float32))))
        loss += self.config["id_coef"] * id_loss
        losses = [actor_loss, critic_loss, id_loss, loss]
        return action, value, action_entropy, loss, losses, sparse_representations

class AKTTaskDiscreteCNNRNN(AKTTaskDiscrete):
    """A3CThread for a discrete action space."""

    def __init__(self, env, task_id, target_task, cluster, monitor_path, config, video=False):
        super(AKTTaskDiscreteCNNRNN, self).__init__(
            env,
            task_id,
            target_task,
            cluster,
            monitor_path,
            config,
            video=video)
        self.initial_features = self.state_init

    def build_shared_network(self):
        x = self.states
        # Convolution layers
        # for i in range(4):
        #     x = tf.nn.elu(conv2d(x, 32, "l{}".format(i + 1), [3, 3], [2, 2]))
        # # Flatten
        # L1 = tf.expand_dims(flatten(x), [0], name="cnns_output")

        # knowledge_base = tf.get_variable("knowledge_base",
        #                                  [self.config["lstm_size"], self.config["n_sparse_units"]],
        #                                  initializer=normalized_columns_initializer(0.01))

        knowledge_bases = []
        n_input_channels = x.get_shape()[3]
        filter_size = (3, 3)
        n_filters = 32
        dtype = tf.float32
        for i in range(4):
            knowledge_base = tf.get_variable(
                "l{}/knowledge_base".format(i + 1),
                [int(filter_size[0] * filter_size[1] * n_input_channels * n_filters) / 4, 4],
                dtype
            )
            print(knowledge_base.get_shape())
            knowledge_bases.append(knowledge_base)
            n_input_channels = n_filters
        return knowledge_bases

    def build_task_network(self, knowledge_bases):
        sparse_representations = []
        x = self.states
        for i in range(4):
            x, sparse = factorized_conv2d(x, knowledge_bases[i], "l{}".format(i + 1), 32, (3, 3), (2, 2))
            x = tf.nn.elu(x)
            sparse_representations.append(sparse)
        x = tf.expand_dims(flatten(x), [0], name="cnns_output")
        lstm_size = self.config["lstm_size"]
        enc_cell = tf.contrib.rnn.BasicLSTMCell(lstm_size)
        lstm_state_size = enc_cell.state_size
        c_init = np.zeros((1, lstm_state_size.c), np.float32)
        h_init = np.zeros((1, lstm_state_size.h), np.float32)
        self.state_init = [c_init, h_init]
        self.rnn_state_in = enc_cell.zero_state(1, tf.float32)
        tf.add_to_collection("rnn_state_in_c", self.rnn_state_in.c)
        tf.add_to_collection("rnn_state_in_h", self.rnn_state_in.h)
        x, self.rnn_state_out = tf.nn.dynamic_rnn(cell=enc_cell,
                                                  inputs=x,
                                                  initial_state=self.rnn_state_in,
                                                  dtype=tf.float32)
        tf.add_to_collection("rnn_state_out_c", self.rnn_state_out.c)
        tf.add_to_collection("rnn_state_out_h", self.rnn_state_out.h)
        x = tf.reshape(x, [-1, lstm_size], name="rnn_output")

        logits = tf.contrib.layers.fully_connected(
            inputs=x,
            num_outputs=self.nA,
            activation_fn=None,
            weights_initializer=normalized_columns_initializer(0.01),
            biases_initializer=tf.zeros_initializer(),
            scope="logits"
        )

        value = tf.contrib.layers.fully_connected(
            inputs=x,
            num_outputs=1,
            activation_fn=None,
            weights_initializer=normalized_columns_initializer(1.0),
            biases_initializer=tf.zeros_initializer(),
            scope="value"
        )
        value = tf.reshape(value, [-1], name="value_output")

        probs = tf.nn.softmax(logits, name="probs")
        action_entropy = probs * tf.nn.log_softmax(logits)

        action = tf.squeeze(tf.multinomial(logits - tf.reduce_max(logits, [1], keepdims=True), 1), [1])
        action = tf.squeeze(tf.one_hot(action, self.nA), name="action")

        actor_loss, critic_loss, loss = actor_critic_discrete_loss(logits,
                                                                   probs,
                                                                   value,
                                                                   self.actions_taken,
                                                                   self.adv,
                                                                   self.r,
                                                                   vf_coef=self.config["vf_coef"],
                                                                   entropy_coef=self.config["entropy_coef"])
        id_loss = 0
        for s in sparse_representations:
            id_loss += tf.reduce_sum(tf.square(s - tf.constant(np.identity(s.get_shape()[0], dtype=np.float32))))
        loss += id_loss
        losses = [actor_loss, critic_loss, id_loss, loss]
        return action, value, action_entropy, loss, losses, sparse_representations

parser = argparse.ArgumentParser()

parser.add_argument("cls", type=str, help="Which class to use for the task.")
parser.add_argument("task_id", type=int, help="Task index.")
parser.add_argument("n_tasks", type=int, help="Total number of tasks in this experiment.")
parser.add_argument("config", type=str, help="Path to config file")
parser.add_argument("--target_task", default=False, action="store_true", help="Say that this is a target task.")
parser.add_argument("-seed", type=int, default=None, help="Seed to use for the environment.")
parser.add_argument("--monitor_path", type=str, help="Path where to save monitor files.")
parser.add_argument("--video", default=False, action="store_true", help="Generate video.")

def main():
    args = parser.parse_args()
    spec = cluster_spec(args.n_tasks, 1)
    cluster = tf.train.ClusterSpec(spec).as_cluster_def()
    cls = load("agents.knowledgetransfer.akt_worker:" + args.cls)
    config = json_to_dict(args.config)
    task = cls(
        make(**config["envs"][args.task_id]),
        args.task_id,
        args.target_task,
        cluster,
        args.monitor_path,
        config,
        video=args.video)
    try:
        task.learn()
    except KeyboardInterrupt:
        pass

if __name__ == '__main__':
    main()



# -*- coding: utf8 -*-

from yarll.misc.utils import load
from yarll.misc.exceptions import ClassNotRegisteredError

# Dictionary with names of algorithms as keys and a list of variants as values.
agent_registry: dict = {}

def register_agent(name: str, entry_point: str, state_dimensions: str, action_space: str, RNN: bool = False):
    """Register an enviroment of a name with a class to be instantiated."""
    if name not in agent_registry:
        agent_registry[name] = [{
            "state_dimensions": state_dimensions,
            "action_space": action_space,
            "RNN": RNN,
            "entry_point": entry_point}]
    else:
        in_list = next((item for item in enumerate(agent_registry[name]) if
                        item[1]["action_space"] == action_space and
                        item[1]["state_dimensions"] == state_dimensions and
                        item[1]["RNN"] == RNN), False)
        if in_list:
            agent_registry[name][in_list[0]]["entry_point"] = entry_point
        else:
            agent_registry[name].append({
                "state_dimensions": state_dimensions,
                "action_space": action_space,
                "RNN": RNN,
                "entry_point": entry_point})

def make_agent(name: str, state_dimensions: str, action_space: str, RNN: bool = False, **args):
    """Make an agent of a given name, possibly using extra arguments."""
    try:
        Agent = agent_registry[name]
        Agent = next((agent_type for agent_type in Agent if
                      agent_type["action_space"] == action_space and
                      agent_type["state_dimensions"] == state_dimensions and
                      agent_type["RNN"] == RNN))
        Agent = Agent["entry_point"]
        if not callable(Agent):
            Agent = load(Agent)
    except (KeyError, StopIteration):
        raise ClassNotRegisteredError(
            "The agent {} for state dimensionality {}, action space {} and RNN={} is not registered.".format(
                name,
                state_dimensions,
                action_space,
                RNN))
    return Agent(**args)



# -*- coding: utf8 -*-

class FunctionApproximator(object):
    """Map states and actions using a function."""
    def __init__(self, n_actions: int) -> None:
        super(FunctionApproximator, self).__init__()
        self.n_actions = n_actions
        self.thetas = []
        self.features_shape = (0)

    def summed_thetas(self, state, action):
        raise NotImplementedError()

    def set_thetas(self, addition):
        self.thetas += addition



# -*- coding: utf8 -*-

import os
import numpy as np
import tensorflow as tf
import logging

from gym import wrappers
# import gym_ple

from yarll.agents.agent import Agent
from yarll.misc.utils import discount_rewards, preprocess_image, FastSaver
from yarll.misc.reporter import Reporter
from yarll.misc.network_ops import create_accumulative_gradients_op, add_accumulative_gradients_op, reset_accumulative_gradients_op

logging.getLogger().setLevel("INFO")

np.set_printoptions(suppress=True)  # Don't use the scientific notation to print results

class KarpathyCNN(Agent):
    """Karpathy policy gradient learner using a convolutional neural network"""
    def __init__(self, env, monitor_path, video=True, **usercfg):
        super(KarpathyCNN, self).__init__(**usercfg)
        self.env = wrappers.Monitor(env, monitor_path, force=True, video_callable=(None if video else False))
        self.nA = env.action_space.n
        self.monitor_path = monitor_path
        # Default configuration. Can be overwritten using keyword arguments.
        self.config.update(
            dict(
                # timesteps_per_batch=10000,
                # n_iter=100,
                n_hidden_units=200,
                learning_rate=1e-3,
                batch_size=10,  # Amount of episodes after which to adapt gradients
                gamma=0.99,  # Discount past rewards by a percentage
                decay=0.99,  # Decay of RMSProp optimizer
                epsilon=1e-9,  # Epsilon of RMSProp optimizer
                draw_frequency=50  # Draw a plot every 50 episodes
            )
        )
        self.config.update(usercfg)
        self.build_network()
        if self.config["save_model"]:
            tf.add_to_collection("action", self.action)
            tf.add_to_collection("states", self.states)
            self.saver = FastSaver()

    def build_network(self):
        image_size = 80
        image_depth = 1  # aka nr. of feature maps. Eg 3 for RGB images. 1 here because we use grayscale images

        self.states = tf.placeholder(tf.float32, [None, image_size, image_size, image_depth], name="states")

        # Convolution layer 1
        depth = 32
        patch_size = 4
        self.w1 = tf.Variable(tf.truncated_normal([patch_size, patch_size, image_depth, depth], stddev=0.01))
        self.b1 = tf.Variable(tf.zeros([depth]))
        self.L1 = tf.nn.relu(tf.nn.conv2d(self.states, self.w1, strides=[1, 2, 2, 1], padding="SAME") + self.b1)
        self.L1 = tf.nn.max_pool(self.L1, ksize=[1, 2, 2, 1], strides=[1, 2, 2, 1], padding="SAME")

        # Convolution layer 2
        self.w2 = tf.Variable(tf.truncated_normal([patch_size, patch_size, depth, depth], stddev=0.01))
        self.b2 = tf.Variable(tf.zeros([depth]))
        self.L2 = tf.nn.relu(tf.nn.conv2d(self.L1, self.w2, strides=[1, 2, 2, 1], padding="SAME") + self.b2)

        # Flatten
        shape = self.L2.get_shape().as_list()
        reshape = tf.reshape(self.L2, [-1, shape[1] * shape[2] * shape[3]])  # -1 for the (unknown) batch size

        # Fully connected layer 1
        self.L3 = tf.contrib.layers.fully_connected(
            inputs=reshape,
            num_outputs=self.config["n_hidden_units"],
            activation_fn=tf.nn.relu,
            weights_initializer=tf.truncated_normal_initializer(mean=0.0, stddev=0.02),
            biases_initializer=tf.zeros_initializer())

        # Fully connected layer 2
        self.probs = tf.contrib.layers.fully_connected(
            inputs=self.L3,
            num_outputs=self.nA,
            activation_fn=tf.nn.softmax,
            weights_initializer=tf.truncated_normal_initializer(mean=0.0, stddev=0.02),
            biases_initializer=tf.zeros_initializer())

        self.action = tf.squeeze(tf.multinomial(tf.log(self.probs), 1), name="action")

        self.vars = [
            self.w1, self.b1,
            self.w2, self.b2,
            self.w3, self.b3,
            self.w4, self.b4
        ]

        self.action_taken = tf.placeholder(tf.float32, shape=[None, self.nA], name="action_taken")
        self.feedback = tf.placeholder(tf.float32, shape=[None, self.nA], name="feedback")
        loss = tf.reduce_mean(tf.squared_difference(self.action_taken, self.probs) * self.feedback)

        self.create_accumulative_grads = create_accumulative_gradients_op(self.vars)
        self.accumulate_grads = add_accumulative_gradients_op(self.vars, self.create_accumulative_grads, loss)
        self.reset_accumulative_grads = reset_accumulative_gradients_op(self.vars, self.create_accumulative_grads)

        self.optimizer = tf.train.RMSPropOptimizer(learning_rate=self.config["learning_rate"], decay=self.config["decay"], epsilon=self.config["epsilon"])

        self.apply_gradients = self.optimizer.apply_gradients(zip(self.create_accumulative_grads, self.vars))

        init = tf.global_variables_initializer()

        # Launch the graph.
        self.session = tf.Session()
        self.session.run(init)

    def choose_action(self, state):
        return self.session.run([self.action], feed_dict={self.states: [state]})[0]

    def get_trajectory(self, render=False):
        """
        Run agent-environment loop for one whole episode (trajectory)
        Return dictionary of results
        Note that this function returns more than the get_trajectory in the EnvRunner class.
        """
        state = preprocess_image(self.env.reset())
        prev_state = state
        states = []
        actions = []
        rewards = []
        for _ in range(self.config["episode_max_length"]):
            delta = state - prev_state
            action = self.choose_action(delta)
            states.append(delta)
            prev_state = state
            state, rew, done, _ = self.env.step(action)
            state = preprocess_image(state)
            actions.append(action)
            rewards.append(rew)
            if done:
                break
            if render:
                self.env.render()
        return {
            "reward": np.array(rewards),
            "state": np.array(states),
            "action": np.array(actions),
        }

    def learn(self):
        reporter = Reporter()

        self.session.run([self.reset_accumulative_grads])

        iteration = 0  # amount of batches processed
        episode_nr = 0
        episode_lengths = np.zeros(self.config["batch_size"])
        episode_rewards = np.zeros(self.config["batch_size"])
        mean_rewards = []
        while True:  # Keep executing episodes
            trajectory = self.get_trajectory()

            episode_rewards[episode_nr % self.config["batch_size"]] = sum(trajectory["reward"])
            episode_lengths[episode_nr % self.config["batch_size"]] = len(trajectory["reward"])
            episode_nr += 1
            action_taken = (np.arange(self.nA) == trajectory["action"][:, None]).astype(np.float32)  # one-hot encoding

            discounted_episode_rewards = discount_rewards(trajectory["reward"], self.config["gamma"])
            # standardize
            discounted_episode_rewards -= np.mean(discounted_episode_rewards)
            std = np.std(discounted_episode_rewards)
            std = std if std > 0 else 1
            discounted_episode_rewards /= std
            feedback = np.reshape(np.repeat(discounted_episode_rewards, self.nA), (len(discounted_episode_rewards), self.nA))

            self.session.run([self.accumulate_grads], feed_dict={self.states: trajectory["state"], self.action_taken: action_taken, self.feedback: feedback})
            if episode_nr % self.config["batch_size"] == 0:  # batch is done
                iteration += 1
                self.session.run([self.apply_gradients])
                self.session.run([self.reset_accumulative_grads])
                reporter.print_iteration_stats(iteration, episode_rewards, episode_lengths, episode_nr)
                mean_rewards.append(episode_rewards.mean())
                if episode_nr % self.config["draw_frequency"] == 0:
                    reporter.draw_rewards(mean_rewards)
        if self.config["save_model"]:
            tf.add_to_collection("action", self.action)
            tf.add_to_collection("states", self.states)
            self.saver.save(self.session, os.path.join(self.monitor_path, "model"))



# -*- coding: utf8 -*-

import gym
from yarll.environment.wrappers import DescriptionWrapper

class Acrobot(DescriptionWrapper):
    """Acrobot-v1 environment wrapper."""

    changeable_parameters = [
        {
            "name": "link_length_1",
            "type": "range",
            "low": 0.2,
            "high": 2.0
        },
        {
            "name": "link_length_2",
            "type": "range",
            "low": 0.2,
            "high": 2.0
        },
        {
            "name": "link_mass_1",
            "type": "range",
            "low": 0.2,
            "high": 2.0
        },
        {
            "name": "link_mass_2",
            "type": "range",
            "low": 0.2,
            "high": 2.0
        },
    ]

    def __init__(self, link_length_1=None, link_length_2=None, link_mass_1=None, link_mass_2=None, **kwargs):
        super(Acrobot, self).__init__(gym.make("OldAcrobot-v1"), **kwargs)
        self.link_length_1 = link_length_1
        self.link_length_2 = link_length_2
        self.link_mass_1 = link_mass_1
        self.link_mass_2 = link_mass_2
        self.change_parameters(
            link_length_1=link_length_1,
            link_length_2=link_length_2,
            link_mass_1=link_mass_1,
            link_mass_2=link_mass_2)

    def change_parameters(self, link_length_1=None, link_length_2=None, link_mass_1=None, link_mass_2=None):
        """Change an Acrobot environment using a different length and/or masspole."""
        if link_length_1 is not None:
            self.unwrapped.LINK_LENGTH_1 = link_length_1
        if link_length_2 is not None:
            self.unwrapped.LINK_LENGTH_2 = link_length_2
        if link_mass_1 is not None:
            self.unwrapped.LINK_MASS_1 = link_mass_1
        if link_mass_2 is not None:
            self.unwrapped.LINK_MASS_2 = link_mass_2



# -*- coding: utf8 -*-

import sys
import os
import tensorflow as tf
from mpi4py import MPI
import numpy as np

from yarll.agents.agent import Agent
from yarll.agents.ppo.ppo import ppo_loss
from yarll.agents.actorcritic.actor_critic import ActorCriticNetworkDiscrete,\
    ActorCriticNetworkDiscreteCNN, ActorCriticNetworkContinuous
from yarll.misc.utils import FastSaver


class DPPO(Agent):
    """Distributed Proximal Policy Optimization agent."""

    RNN = False

    def __init__(self, env, monitor_path, **usercfg):
        super(DPPO, self).__init__()
        self.env = env
        self.env_name: str = env.spec.id
        self.monitor_path: str = monitor_path

        self.comm = MPI.COMM_SELF

        self.config.update(dict(
            n_workers=3,
            n_hidden_units=20,
            n_hidden_layers=2,
            gamma=0.99,
            gae_lambda=0.95,
            learning_rate=2.5e-4,
            n_iter=10000,
            n_epochs=4,
            n_local_steps=128,
            gradient_clip_value=0.5,
            vf_coef=0.5,
            entropy_coef=0.01,
            cso_epsilon=0.1,  # Clipped surrogate objective epsilon
            learn_method="batches",
            batch_size=64,
            save_model=False
        ))
        self.config.update(usercfg)

        self.task_type = None  # To be filled in by subclasses

        self.n_updates: int = 0

        with tf.variable_scope("new_network"):
            self.new_network = self.build_networks()
            if self.RNN:
                self.initial_features = self.new_network.state_init
            else:
                self.initial_features = None
            self.new_network_vars = tf.get_collection(
                tf.GraphKeys.TRAINABLE_VARIABLES, tf.get_variable_scope().name)
        self._global_step = tf.get_variable(
            "global_step",
            [],
            tf.int32,
            initializer=tf.constant_initializer(0, dtype=tf.int32),
            trainable=False)
        self.action = self.new_network.action
        self.value = self.new_network.value
        self.states = self.new_network.states
        self.actions_taken = self.new_network.actions_taken
        self.advantage = tf.placeholder(tf.float32, [None], name="advantage")
        self.ret = tf.placeholder(tf.float32, [None], name="return")

        with tf.variable_scope("old_network"):
            self.old_network = self.build_networks()
            self.old_network_vars = tf.get_collection(
                tf.GraphKeys.TRAINABLE_VARIABLES, tf.get_variable_scope().name)

        self.set_old_to_new = tf.group(
            *[v1.assign(v2) for v1, v2 in zip(self.old_network_vars, self.new_network_vars)])

        # Reduces by taking the mean instead of summing
        self.actor_loss = -tf.reduce_mean(self.make_actor_loss(self.old_network, self.new_network, self.advantage))
        self.critic_loss = tf.reduce_mean(tf.square(self.value - self.ret))
        self.mean_entropy = tf.reduce_mean(self.new_network.entropy)
        self.loss = self.actor_loss + self.config["vf_coef"] * self.critic_loss + \
            self.config["entropy_coef"] * self.mean_entropy

        grads = tf.gradients(self.loss, self.new_network_vars)

        self.n_steps = tf.shape(self.states)[0]
        if self.config["save_model"]:
            tf.add_to_collection("action", self.action)
            tf.add_to_collection("states", self.states)
            self.saver = FastSaver()
        summary_actor_loss = tf.summary.scalar(
            "model/Actor_loss", self.actor_loss)
        summary_critic_loss = tf.summary.scalar(
            "model/Critic_loss", self.critic_loss)
        summary_loss = tf.summary.scalar("model/Loss", self.loss)
        summary_entropy = tf.summary.scalar("model/Entropy", -self.mean_entropy)
        summary_grad_norm = tf.summary.scalar(
            "model/grad_global_norm", tf.global_norm(grads))
        summary_var_norm = tf.summary.scalar(
            "model/var_global_norm", tf.global_norm(self.new_network_vars))
        self.model_summary_op = tf.summary.merge([
            summary_actor_loss,
            summary_critic_loss,
            summary_loss,
            summary_entropy,
            summary_grad_norm,
            summary_var_norm
        ])
        self.writer = tf.summary.FileWriter(os.path.join(
            self.monitor_path, "master"))

        # grads before clipping were passed to the summary, now clip and apply them
        if self.config["gradient_clip_value"] is not None:
            grads, _ = tf.clip_by_global_norm(grads, self.config["gradient_clip_value"])

        with tf.variable_scope("optimizer"):
            self.optimizer = tf.train.AdamOptimizer(
                self.config["learning_rate"], name="optim")
            apply_grads = self.optimizer.apply_gradients(
                zip(grads, self.new_network_vars))

            inc_step = self._global_step.assign_add(self.n_steps)
            self.train_op = tf.group(apply_grads, inc_step)
        optimizer_variables = [var for var in tf.global_variables() if var.name.startswith("optimizer")]
        self.init_op = tf.variables_initializer(self.new_network_vars + optimizer_variables + [self._global_step])

    def make_actor_loss(self, old_network, new_network, advantage):
        return ppo_loss(old_network.action_log_prob, new_network.action_log_prob, self.config["cso_epsilon"], advantage)

    def build_networks(self):
        raise NotImplementedError

    def update_network(self, states, actions, advs, returns, features=None):
        fetches = [self.model_summary_op, self.train_op]
        feed_dict = {
            self.states: states,
            self.old_network.states: states,
            self.actions_taken: actions,
            self.old_network.actions_taken: actions,
            self.advantage: advs,
            self.ret: returns
        }
        if features != [] and features is not None:
            feed_dict[self.old_network.rnn_state_in] = features
            feed_dict[self.new_network.rnn_state_in] = features
        summary, _ = tf.get_default_session().run(fetches, feed_dict)
        self.writer.add_summary(summary, self.n_updates)
        self.n_updates += 1

    def learn_by_batches(self, trajectories):
        all_states, all_actions, all_advs, all_returns = [], [], [], []
        for states, actions, advs, returns, _ in trajectories:
            all_states.extend(states)
            all_actions.extend(actions)
            all_advs.extend(advs)
            all_returns.extend(returns)
        all_advs = np.array(all_advs)
        all_advs = (all_advs - all_advs.mean()) / all_advs.std()
        indices = np.arange(len(all_states))
        batch_size = int(self.config["batch_size"])
        for _ in range(int(self.config["n_epochs"])):
            np.random.shuffle(indices)
            for j in range(0, len(all_states), batch_size):
                batch_indices = indices[j:(j + batch_size)]
                batch_states = np.array(all_states)[batch_indices]
                batch_actions = np.array(all_actions)[batch_indices]
                batch_advs = np.array(all_advs)[batch_indices]
                batch_rs = np.array(all_returns)[batch_indices]
                self.update_network(batch_states, batch_actions, batch_advs, batch_rs)
            self.writer.flush()


    def learn_by_trajectories(self, trajectories):
        for _ in range(int(self.config["n_epochs"])):
            for states, actions, advs, returns, features in trajectories:
                self.update_network(states, actions, advs, returns, features)
            self.writer.flush()

    def learn(self):
        """Run learning algorithm"""
        config = self.config
        current_folder = os.path.abspath(
            os.path.dirname(os.path.realpath(__file__)))
        args = [
            os.path.join(current_folder, "dppo_worker.py"),
            self.env_name,
            self.task_type,
            self.config["config_path"],
            "--monitor_path", self.monitor_path
        ]
        seed = self.config["seed"]
        if seed is not None:
            args += ["--seed", str(seed)]
        comm = self.comm.Spawn(
            sys.executable,
            args=args,
            maxprocs=int(self.config["n_workers"])
        )
        sess_config = tf.ConfigProto()
        sess_config.gpu_options.allow_growth = True
        with tf.Session(config=sess_config) as sess, sess.as_default():
            tf.get_default_session().run(self.init_op)
            for _ in range(config["n_iter"]):
                # Collect trajectories until we get timesteps_per_batch total timesteps
                for var in self.new_network_vars:
                    comm.Bcast(var.eval(), root=MPI.ROOT)
                trajectories = comm.gather(None, root=MPI.ROOT)
                tf.get_default_session().run(self.set_old_to_new)

                # Mix steps of all trajectories and learn by minibatches or not
                if self.config["learn_method"] == "batches":
                    self.learn_by_batches(trajectories)
                else:
                    self.learn_by_trajectories(trajectories)

class DPPODiscrete(DPPO):

    def __init__(self, env, monitor_path, **usercfg):
        super(DPPODiscrete, self).__init__(env, monitor_path, **usercfg)
        self.task_type = "DPPOWorkerDiscrete"

    def build_networks(self):
        return ActorCriticNetworkDiscrete(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]),
            int(self.config["n_hidden_layers"]))


class DPPODiscreteCNN(DPPODiscrete):

    def __init__(self, env, monitor_path, **usercfg):
        super(DPPODiscreteCNN, self).__init__(env, monitor_path, **usercfg)
        self.task_type = "DPPOWorkerDiscreteCNN"

    def build_networks(self):
        return ActorCriticNetworkDiscreteCNN(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]))


class DPPOContinuous(DPPO):

    def __init__(self, env, monitor_path, **usercfg):
        super(DPPOContinuous, self).__init__(env, monitor_path, **usercfg)
        self.task_type = "DPPOWorkerContinuous"

    def build_networks(self):
        return ActorCriticNetworkContinuous(
            list(self.env.observation_space.shape),
            self.env.action_space,
            int(self.config["n_hidden_units"]),
            int(self.config["n_hidden_layers"]))

    def get_env_action(self, action):
        return action



#!/usr/bin/env python
# # -*- coding: utf8 -*-

import argparse
import os
from typing import Dict, Any
import numpy as np
from mpi4py import MPI
os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"   # see issue #152
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
import tensorflow as tf

from yarll.environment.registration import make
from yarll.misc.utils import load, json_to_dict
from yarll.agents.actorcritic.actor_critic import ActorCriticNetworkDiscrete, ActorCriticNetworkDiscreteCNN, ActorCriticNetworkDiscreteCNNRNN, actor_critic_discrete_loss, ActorCriticNetworkContinuous, actor_critic_continuous_loss
from yarll.agents.env_runner import EnvRunner


class DPPOWorker(object):
    """Distributed Proximal Policy Optimization Worker."""

    def __init__(self, env_id: str, task_id: int, comm: MPI.Intercomm, monitor_path: str, config: Dict[str, Any], seed=None) -> None:
        super(DPPOWorker, self).__init__()
        self.comm = comm
        self.config = config
        self.env = make(env_id)
        self.task_id = task_id
        if seed is not None:
            self.env.seed(seed)
        self.writer = tf.summary.FileWriter(os.path.join(
            monitor_path,
            "task{}".format(task_id)))

        # Only used (and overwritten) by agents that use an RNN
        self.initial_features = None
        with tf.device('/cpu:0'):
            with tf.variable_scope("new_network"):  # The workers only have 1 network
                self.global_network = self.build_networks()
                self.states = self.global_network.states
                self.action = self.global_network.action
                self.value = self.global_network.value
                self.global_vars = tf.get_collection(
                    tf.GraphKeys.TRAINABLE_VARIABLES, tf.get_variable_scope().name)
                self._global_step = tf.get_variable(
                    "global_step", [],
                    tf.int32,
                    initializer=tf.constant_initializer(0, dtype=tf.int32),
                    trainable=False)

            self.env_runner = EnvRunner(
                self.env, self, {}, summary_writer=self.writer)

    def build_networks(self):
        raise NotImplementedError

    def run(self):
        with tf.Session() as sess, sess.as_default():
            var_receivers = [np.zeros(var.shape.as_list(), dtype=var.dtype.as_numpy_dtype) for var in self.global_vars]
            while True:
                for var_receiver, tf_var in zip(var_receivers, self.global_vars):
                    self.comm.Bcast(var_receiver, root=0)
                    tf_var.load(var_receiver)
                experiences = self.env_runner.get_steps(
                    int(self.config["n_local_steps"]), stop_at_trajectory_end=False)
                T = experiences.steps
                value = 0 if experiences.terminals[-1] else self.get_critic_value(
                    np.asarray(experiences.states)[None, -1], experiences.features[-1])
                vpred = np.asarray(experiences.values + [value])
                gamma = self.config["gamma"]
                lambda_ = self.config["gae_lambda"]
                gaelam = advantages = np.empty(T, 'float32')
                last_gaelam = 0
                for t in reversed(range(T)):
                    nonterminal = 1 - experiences.terminals[t]
                    delta = experiences.rewards[t] + gamma * vpred[t + 1] * nonterminal - vpred[t]
                    gaelam[t] = last_gaelam = delta + gamma * lambda_ * nonterminal * last_gaelam
                returns = advantages + experiences.values
                processed = experiences.states, experiences.actions, advantages, returns, experiences.features[0]
                self.comm.gather(processed, root=0)

    @property
    def global_step(self):
        return self._global_step.eval()

    def get_critic_value(self, state, *rest):
        fetches = [self.global_network.value]
        feed_dict = {self.global_network.states: state}
        value = tf.get_default_session().run(fetches, feed_dict=feed_dict)[0].flatten()
        return value

    def choose_action(self, state, *rest):
        fetches = [self.global_network.action, self.global_network.value]
        feed_dict = {
            self.global_network.states: [state]
        }
        action, value = tf.get_default_session().run(fetches, feed_dict=feed_dict)
        return {"action": action, "value": value[0]}

    def get_env_action(self, action):
        return np.argmax(action)

    def new_trajectory(self):
        pass


class DPPOWorkerDiscrete(DPPOWorker):
    """DPPOWorker for a discrete action space."""

    def __init__(self, env_id, task_id, comm, monitor_path, config, seed=None):
        self.make_loss = actor_critic_discrete_loss
        super(DPPOWorkerDiscrete, self).__init__(
            env_id,
            task_id,
            comm,
            monitor_path,
            config,
            seed
        )

    def build_networks(self):
        ac_net = ActorCriticNetworkDiscrete(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]),
            int(self.config["n_hidden_layers"]))
        return ac_net


class DPPOWorkerDiscreteCNN(DPPOWorkerDiscrete):
    """DPPOWorker for a discrete action space."""

    def __init__(self, env_id, task_id, comm, monitor_path, config, seed=None):
        self.make_loss = actor_critic_discrete_loss
        super(DPPOWorkerDiscreteCNN, self).__init__(
            env_id,
            task_id,
            comm,
            monitor_path,
            config,
            seed
        )

    def build_networks(self):
        ac_net = ActorCriticNetworkDiscreteCNN(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]),
            summary=False)
        return ac_net


class DPPOWorkerDiscreteCNNRNN(DPPOWorkerDiscreteCNN):
    """DPPOWorker for a discrete action space."""

    def __init__(self, env_id, task_id, comm, monitor_path, config, seed=None):
        self.make_loss = actor_critic_discrete_loss
        super(DPPOWorkerDiscreteCNNRNN, self).__init__(
            env_id,
            task_id,
            comm,
            monitor_path,
            config,
            seed
        )

    def build_networks(self):
        ac_net = ActorCriticNetworkDiscreteCNNRNN(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]),
            summary=False)
        self.initial_features = ac_net.state_init
        return ac_net

    def choose_action(self, state, features):
        feed_dict = {
            self.states: [state]
        }

        feed_dict[self.global_network.rnn_state_in] = features
        action, rnn_state, value = tf.get_default_session().run(
            [self.action, self.global_network.rnn_state_out, self.value],
            feed_dict=feed_dict)
        return {"action": action, "value": value, "features": rnn_state}

    def get_critic_value(self, states, features):
        feed_dict = {
            self.states: states
        }
        feed_dict[self.global_network.rnn_state_in] = features
        return tf.get_default_session().run(self.value, feed_dict=feed_dict)[0]


class DPPOWorkerContinuous(DPPOWorker):
    """DPPOWorker for a continuous action space."""

    def __init__(self, env_id, task_id, comm, monitor_path, config, seed=None):
        self.make_loss = actor_critic_continuous_loss
        super(DPPOWorkerContinuous, self).__init__(
            env_id,
            task_id,
            comm,
            monitor_path,
            config,
            seed
        )

    def build_networks(self):
        ac_net = ActorCriticNetworkContinuous(
            list(self.env.observation_space.shape),
            self.env.action_space,
            int(self.config["n_hidden_units"]),
            int(self.config["n_hidden_layers"]))
        return ac_net

    def get_env_action(self, action):
        return action


parser = argparse.ArgumentParser()

parser.add_argument("env_id", type=str,
                    help="Name of the environment on which to run")
parser.add_argument("cls", type=str, help="Which class to use for the task.")
parser.add_argument("config", type=str, help="Path to config file")
parser.add_argument("-seed", type=int, default=None,
                    help="Seed to use for the environment.")
parser.add_argument("--monitor_path", type=str,
                    help="Path where to save monitor files.")
parser.add_argument("--seed", type=int, default=None, help="Seed to use for environments.")


def main():
    comm = MPI.Comm.Get_parent()
    task_id = comm.Get_rank()
    args = parser.parse_args()
    cls = load("agents.ppo.dppo_worker:" + args.cls)
    config = json_to_dict(args.config)

    task = cls(args.env_id, task_id, comm, args.monitor_path, config, args.seed)
    task.run()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import os
import queue
import threading
import argparse
import time
from typing import Optional
import tensorflow as tf
import numpy as np
from gym import wrappers

from yarll.environment.registration import make
from yarll.misc.network_ops import create_sync_net_op
from yarll.misc.utils import discount_rewards, FastSaver, load, json_to_dict, cluster_spec
from yarll.agents.actorcritic.actor_critic import ActorCriticNetworkDiscrete, ActorCriticNetworkDiscreteCNN, \
ActorCriticNetworkDiscreteCNNRNN, actor_critic_discrete_loss, ActorCriticNetworkContinuous, actor_critic_continuous_loss
from yarll.memory.experiences_memory import ExperiencesMemory

def env_runner(env, policy, n_steps: int, render=False, summary_writer=None):
    """
    Run agent-environment loop for maximally n_steps.
    Yields a dictionary of results.
    """
    episode_steps = 0
    episode_reward = 0
    n_episodes = 0
    state = env.reset()
    features = policy.initial_features
    timestep_limit = env.spec.tags.get('wrapper_config.TimeLimit.max_episode_steps')

    while True:
        memory = ExperiencesMemory()

        for _ in range(n_steps):
            # Choose the next action (using a neural network) depending on the current state
            results = policy.choose_action(state, features)
            action = results["action"]
            value = results.get("value", None)
            new_features = results.get("features", None)
            # Execute the action in the environment
            new_state, reward, terminal, _ = env.step(policy.get_env_action(action))
            episode_steps += 1
            episode_reward += reward
            memory.add(state, action, reward, value, features, terminal)
            state = new_state
            features = new_features
            if terminal or episode_steps >= timestep_limit:
                n_episodes += 1
                features = policy.initial_features
                if episode_steps >= timestep_limit or not env.metadata.get('semantics.autoreset'):
                    state = env.reset()
                if summary_writer is not None:
                    summary = tf.Summary()
                    summary.value.add(tag="global/Episode_length", simple_value=float(episode_steps))
                    summary.value.add(tag="global/Reward", simple_value=float(episode_reward))
                    summary_writer.add_summary(summary, n_episodes)
                    summary_writer.flush()
                episode_steps = 0
                episode_reward = 0
                break
            if render:
                env.render()
        yield memory


class RunnerThread(threading.Thread):
    """
    Thread that collects trajectories from the environment
    and puts them on a queue.
    """

    def __init__(self, env, policy, n_local_steps: int, render=False) -> None:
        super(RunnerThread, self).__init__()
        self.env = env
        self.policy = policy
        self.n_local_steps = n_local_steps
        self.render = render
        self.daemon = True
        self.sess = None
        self.summary_writer = None
        self.queue: queue.Queue = queue.Queue(maxsize=5)

    def start_runner(self, sess, summary_writer):
        self.sess = sess
        self.summary_writer = summary_writer
        self.start()

    def run(self):
        trajectory_provider = env_runner(self.env, self.policy, self.n_local_steps, self.render, self.summary_writer)
        while True:
            # the timeout variable exists because apparently, if one worker dies, the other workers
            # won't die with it, unless the timeout is set to some large number.
            # This is an empirical observation.
            self.queue.put(next(trajectory_provider), timeout=600.0)


class A3CTask(object):
    """Single A3C learner thread."""

    def __init__(self,
                 env_id: str,
                 task_id: int,
                 cluster: tf.train.ClusterDef,
                 monitor_path: str,
                 config: dict,
                 clip_gradients: bool = True,
                 video: bool = False,
                 seed: Optional[int] = None) -> None:
        super(A3CTask, self).__init__()
        self.task_id = task_id
        self.config = config
        self.clip_gradients = clip_gradients
        self.env = make(env_id)
        self.env.seed(seed)
        if task_id == 0:
            self.env = wrappers.Monitor(
                self.env,
                monitor_path,
                force=True,
                video_callable=(None if video else False)
            )

        # Only used (and overwritten) by agents that use an RNN
        self.initial_features = None

        worker_device = "/job:worker/task:{}/cpu:0".format(task_id)
        # Global network
        shared_device = tf.train.replica_device_setter(
            ps_tasks=1,
            worker_device=worker_device,
            cluster=cluster)
        with tf.device(shared_device):
            with tf.variable_scope("global"):
                self.global_network = self.build_networks()
                self.global_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, tf.get_variable_scope().name)
                self._global_step = tf.get_variable(
                    "global_step",
                    [],
                    tf.int32,
                    initializer=tf.constant_initializer(0, dtype=tf.int32),
                    trainable=False)

        # Local network
        with tf.device(worker_device):
            with tf.variable_scope("local"):
                self.local_network = self.build_networks()
                self.states = self.local_network.states
                self.actions_taken = self.local_network.actions_taken
                self.advantage = tf.placeholder(tf.float32, [None], name="advantage")
                self.ret = tf.placeholder(tf.float32, [None], name="return")
                self.actor_loss, self.critic_loss, self.loss = self.make_loss()
                self.local_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, tf.get_variable_scope().name)
                self.sync_net = create_sync_net_op(self.global_vars, self.local_vars)
                self.n_steps = tf.shape(self.local_network.states)[0]
                inc_step = self._global_step.assign_add(self.n_steps)

        device = shared_device if self.config["shared_optimizer"] else worker_device
        with tf.device(device):
            apply_optim_op = self.make_trainer()
            self.train_op = tf.group(apply_optim_op, inc_step)

            loss_summaries = self.create_summary_losses()
            self.reward = tf.placeholder("float", name="reward")
            tf.summary.scalar("Reward", self.reward)
            self.episode_length = tf.placeholder("float", name="episode_length")
            tf.summary.scalar("Episode_length", self.episode_length)
            self.summary_op = tf.summary.merge(loss_summaries)

        variables_to_save = [v for v in tf.global_variables() if not v.name.startswith("local")]
        init_op = tf.variables_initializer(variables_to_save)
        init_all_op = tf.global_variables_initializer()
        saver = FastSaver(variables_to_save)
        # Write the summary of each task in a different directory
        self.writer = tf.summary.FileWriter(os.path.join(monitor_path, "task{}".format(task_id)))

        self.runner = RunnerThread(self.env, self, int(self.config["n_local_steps"]), task_id == 0 and video)

        self.server = tf.train.Server(
            cluster,
            job_name="worker",
            task_index=task_id,
            config=tf.ConfigProto(intra_op_parallelism_threads=1, inter_op_parallelism_threads=2)
        )

        def init_fn(scaffold, sess):
            sess.run(init_all_op)

        self.report_uninit_op = tf.report_uninitialized_variables(variables_to_save)

        self.scaffold = tf.train.Scaffold(
            init_op=init_op,
            init_fn=init_fn,
            ready_for_local_init_op=self.report_uninit_op,
            saver=saver,
            ready_op=self.report_uninit_op
        )

        self.config_proto = tf.ConfigProto(device_filters=["/job:ps", "/job:worker/task:{}/cpu:0".format(task_id)])

        self.session = None

    def build_networks(self):
        raise NotImplementedError()

    def make_loss(self):
        raise NotImplementedError()

    def get_critic_value(self, states, features):
        return self.session.run(self.local_network.value, feed_dict={self.states: states})[0]

    def get_env_action(self, action):
        return np.argmax(action)

    def choose_action(self, state, features):
        """Choose an action."""
        feed_dict = {
            self.states: [state]
        }
        action, value = self.session.run([self.local_network.action, self.local_network.value],
                                         feed_dict=feed_dict)
        return {"action": action, "value": value}

    def make_trainer(self):
        optimizer = tf.train.AdamOptimizer(self.config["learning_rate"], name="optim")
        grads = tf.gradients(self.loss, self.local_vars)
        grads, _ = tf.clip_by_global_norm(grads, self.config["gradient_clip_value"])

        # Apply gradients to the weights of the master network
        return optimizer.apply_gradients(zip(grads, self.global_vars))

    def create_summary_losses(self):
        n_steps = tf.to_float(self.n_steps)
        actor_loss_summary = tf.summary.scalar("model/actor_loss", tf.squeeze(self.actor_loss / n_steps))
        critic_loss_summary = tf.summary.scalar("model/critic_loss", tf.squeeze(self.critic_loss / n_steps))
        loss_summary = tf.summary.scalar("model/loss", tf.squeeze(self.loss / n_steps))
        return [actor_loss_summary, critic_loss_summary, loss_summary]

    def pull_batch_from_queue(self):
        """
        Take a trajectory from the queue.
        Also immediately try to extend it if the episode
        wasn't over and more transitions are available.
        """
        trajectory = self.runner.queue.get(timeout=600.0)
        while not trajectory.terminal:
            try:
                trajectory.extend(self.runner.queue.get_nowait())
            except queue.Empty:
                break
        return trajectory

    @property
    def global_step(self):
        return self._global_step.eval(session=self.session)

    def learn(self):
        # Assume global shared parameter vectors θ and θv and global shared counter T = 0
        # Assume thread-specific parameter vectors θ' and θ'v
        if self.task_id != 0:
            time.sleep(5)
        with tf.train.MonitoredTrainingSession(
            master=self.server.target,
            is_chief=(self.task_id == 0),
            config=self.config_proto,
            save_summaries_secs=30,
            scaffold=self.scaffold
        ) as sess:
            self.session = sess
            sess.run(self.sync_net)
            self.runner.start_runner(sess, self.writer)
            while not sess.should_stop() and self.global_step < self.config["T_max"]:
                # Synchronize thread-specific parameters θ' = θ and θ'v = θv
                sess.run(self.sync_net)
                trajectory = self.pull_batch_from_queue()
                v = 0 if trajectory.terminal else self.get_critic_value(
                    np.asarray(trajectory.states)[None, -1], trajectory.features[-1])
                rewards_plus_v = np.asarray(trajectory.rewards + [v])
                vpred_t = np.asarray(trajectory.values + [v])
                delta_t = trajectory.rewards + self.config["gamma"] * vpred_t[1:] - vpred_t[:-1]
                batch_r = discount_rewards(rewards_plus_v, self.config["gamma"])[:-1]
                batch_adv = discount_rewards(delta_t, self.config["gamma"])
                fetches = [self.summary_op, self.train_op, self._global_step]
                states = np.asarray(trajectory.states)
                feed_dict = {
                    self.states: states,
                    self.actions_taken: np.asarray(trajectory.actions),
                    self.advantage: batch_adv,
                    self.ret: np.asarray(batch_r)
                }
                feature = trajectory.features[0]
                if feature != [] and feature is not None:
                    feed_dict[self.local_network.rnn_state_in] = feature
                summary, _, global_step = sess.run(fetches, feed_dict)
                self.writer.add_summary(summary, global_step)
                self.writer.flush()


class A3CTaskDiscrete(A3CTask):
    """A3CTask for a discrete action space."""

    def build_networks(self):
        ac_net = ActorCriticNetworkDiscrete(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]),
            int(self.config["n_hidden_layers"]))
        return ac_net

    def make_loss(self):
        return actor_critic_discrete_loss(
            self.local_network.logits,
            self.local_network.probs,
            self.local_network.value,
            self.local_network.actions_taken,
            self.advantage,
            self.ret,
            self.config["vf_coef"],
            self.config["entropy_coef"],
            self.config["loss_reducer"]
        )


class A3CTaskDiscreteCNN(A3CTaskDiscrete):
    """A3CTask for a discrete action space."""

    def build_networks(self):
        ac_net = ActorCriticNetworkDiscreteCNN(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]),
            summary=False)
        return ac_net


class A3CTaskDiscreteCNNRNN(A3CTaskDiscreteCNN):
    """A3CTask for a discrete action space."""

    def build_networks(self):
        ac_net = ActorCriticNetworkDiscreteCNNRNN(
            list(self.env.observation_space.shape),
            self.env.action_space.n,
            int(self.config["n_hidden_units"]),
            summary=False)
        self.initial_features = ac_net.state_init
        return ac_net

    def choose_action(self, state, features):
        feed_dict = {
            self.local_network.states: [state],
            self.local_network.rnn_state_in: features
        }
        action, rnn_state, value = self.session.run(
            [self.local_network.action, self.local_network.rnn_state_out, self.local_network.value],
            feed_dict=feed_dict)
        return {"action": action, "value": value, "features": rnn_state}

    def get_critic_value(self, states, features):
        feed_dict = {
            self.local_network.states: states,
            self.local_network.rnn_state_in: features
        }
        return self.session.run(self.local_network.value, feed_dict=feed_dict)[0]


class A3CTaskContinuous(A3CTask):
    """A3CTask for a continuous action space."""

    def build_networks(self):
        ac_net = ActorCriticNetworkContinuous(
            list(self.env.observation_space.shape),
            self.env.action_space,
            int(self.config["n_hidden_units"]),
            int(self.config["n_hidden_layers"]))
        return ac_net

    def make_loss(self):
        return actor_critic_continuous_loss(
            self.local_network.action_log_prob,
            self.local_network.entropy,
            self.local_network.value,
            self.advantage,
            self.ret,
            self.config["vf_coef"],
            self.config["entropy_coef"],
            self.config["loss_reducer"]
        )

    def get_env_action(self, action):
        return action


parser = argparse.ArgumentParser()

parser.add_argument("env_id", type=str, help="Name of the environment on which to run")
parser.add_argument("cls", type=str, help="Which class to use for the task.")
parser.add_argument("task_id", type=int, help="Task index.")
parser.add_argument("n_tasks", type=int, help="Total number of tasks in this experiment.")
parser.add_argument("config", type=str, help="Path to config file")
parser.add_argument("-seed", type=int, default=None, help="Seed to use for the environment.")
parser.add_argument("--monitor_path", type=str, help="Path where to save monitor files.")
parser.add_argument("--video", default=False, action="store_true", help="Generate video.")

def main():
    args = parser.parse_args()
    spec = cluster_spec(args.n_tasks, 1)
    cluster = tf.train.ClusterSpec(spec).as_cluster_def()
    cls = load("agents.actorcritic.a3c_worker:" + args.cls)
    config = json_to_dict(args.config)
    task = cls(args.env_id, args.task_id, cluster, args.monitor_path, config, video=args.video)
    task.learn()

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf8 -*-

import os
import re
import json
import argparse
import operator
import logging
import numpy as np
from tensorboard.backend.event_processing.plugin_event_multiplexer import EventMultiplexer
import matplotlib.pyplot as plt

from yarll.misc.utils import ge
from yarll.misc.exceptions import WrongArgumentsError

IMAGES_EXT = ".png"

logging.getLogger("tensorflow").setLevel(logging.WARNING)

def moving_average(a, n):
    """
    Compute the moving average of an array a of numbers using a window length n.
    Source: http://stackoverflow.com/questions/14313510/how-to-calculate-moving-average-using-numpy
    """
    ret = np.cumsum(a, dtype=float)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n

# Source: https://github.com/tensorflow/tensorflow/blob/d413f62fba038d820b775d95ddd518fb6f43d3ed/tensorflow/tensorboard/components/vz_line_chart/vz-line-chart.ts#L432
def exponential_smoothing(a, weight):
    factor = (pow(1000, weight) - 1) / 999
    length = len(a)
    kernelRadius = np.floor(length * factor / 2)
    result = []
    for i, x in enumerate(a):
        actualKernelRadius = min(kernelRadius, i)
        start = int(i - actualKernelRadius)
        end = int(i + actualKernelRadius + 1)
        if end >= length:
            result.append(np.inf)
        elif x == np.inf:
            result.append(x)
        else:
            result.append(np.mean(list(filter(lambda x: x != np.inf, a[start:end]))))
    return result

def create_smoother(f, *args):
    return lambda x: f(x, *args)

def plot(x, y, x_label: str, scalar, xmax=None, ymin=None, ymax=None):
    x_label_upper = x_label[0].upper() + x_label[1:]
    fig = plt.figure()
    plt.plot(x, y)
    plt.xlim(xmax=xmax)
    plt.ylim(ymin=ymin, ymax=ymax)
    plt.xlabel(x_label_upper)
    plt.ylabel(scalar)
    plt.title("{} per {}".format(scalar, x_label))
    fig.canvas.set_window_title("{} per {}".format(scalar, x_label))

def plot_tasks(data, x_label: str, smoothing_function=None, xmin=None, xmax=None, max_reward=None, legend=True, save_directory=None, show_plots=True):
    x_label_upper = x_label[0].upper() + x_label[1:]
    for scalar, tasks in data.items():
        fig = plt.figure()
        for task, epochs_values in sorted(tasks.items(), key=operator.itemgetter(0)):
            mean = np.mean(epochs_values["values"], axis=0)
            if smoothing_function:
                mean = smoothing_function(mean)
            # percentiles = np.percentile(epochs_values["values"], [25, 75], axis=0)
            std = np.std(epochs_values["values"], axis=0)
            std = std[len(std) - len(mean):]
            # error_min, error_max = mean - std, mean + std
            plt.plot(epochs_values["epochs"], mean, label="Task " + str(task))
            # plt.fill_between(x, error_min, error_max, alpha=0.3)
            # min_y = min(min_y, min(error_min))
            # max_y = max(max_y, max(filter(lambda x: x != np.inf, error_max)))
        if legend:
            plt.legend()
        plt.xlim(xmin=xmin, xmax=xmax)
        # plt.ylim(ymin=min(0, min_y), ymax=max(0, max_y + 0.1 * max_y))
        if "reward" in scalar.lower() and max_reward is not None:
            ymax = max_reward * 1.1
        else:
            ymax = None
        plt.ylim(ymax=ymax)
        plt.xlabel(x_label_upper)
        plt.ylabel(scalar)
        plt.title("{} per {}".format(scalar, x_label))
        fig.canvas.set_window_title("{} per {}".format(scalar, x_label))
        if save_directory is not None:
            plt.savefig(os.path.join(save_directory, "{}_per_{}".format(scalar, x_label) + IMAGES_EXT))
    if show_plots:
        plt.show()

def plot_gym_monitor_stats(stats_path, xmax=None, smoothing_function=None, save_directory=None, show_plots=True):
    f = open(stats_path)
    contents = json.load(f)
    data = contents["episode_rewards"]
    if smoothing_function:
        data = smoothing_function(data)
    min_aer = min(data)
    max_aer = max(filter(lambda x: x != np.inf, data))
    ymin = min(0, min_aer)
    ymax = max(0, max_aer + 0.1 * max_aer)
    plot(range(len(data)), data, "episode", "total reward", xmax=xmax, ymin=ymin, ymax=ymax)
    if save_directory is not None:
        plt.savefig(os.path.join(save_directory, "totalreward_per_episode" + IMAGES_EXT))

    data = contents["episode_lengths"]
    if smoothing_function:
        data = smoothing_function(data)
    max_ael = max(filter(lambda x: x != np.inf, data))
    ymin = 0
    ymax = (max_ael + 0.1 * max_ael)
    plot(range(len(data)), data, "episode", "length", xmax=xmax, ymin=ymin, ymax=ymax)
    if save_directory is not None:
        plt.savefig(os.path.join(save_directory, "length_per_episode" + IMAGES_EXT))
    if show_plots:
        plt.show()

def tf_scalar_data(em):
    """Process scalar data of TensorFlow summaries."""
    runs = list(em.Runs().keys())
    scalars = em.Runs()[runs[0]]["tensors"]  # Assumes that the scalars of every run are the same
    pattern = re.compile("[a-zA-Z]+(\d+)$")
    data = {}
    for scalar in scalars:
        scalar_data = {}
        for run in runs:
            task = int(pattern.search(run).group(1))
            accum = em.GetAccumulator(run)
            if task in scalar_data:
                scalar_data[task]["values"].append([s.tensor_proto.float_val[0] for s in accum.Tensors(scalar)])
            else:
                values = []
                epochs = []
                for s in accum.Tensors(scalar):
                    values.append(s.tensor_proto.float_val[0])
                    epochs.append(s.step)
                scalar_data[task] = {
                    "epochs": epochs,
                    "values": [values]
                }
        data[scalar] = scalar_data
    return data

def plot_tf_scalar_summaries(summaries_dir,
                             xmin=None,
                             xmax=None,
                             smoothing_function=None,
                             max_reward=None,
                             x_label="episode",
                             legend=True,
                             save_directory=None,
                             show_plots=True):
    """Plot TensorFlow scalar summaries."""
    em = EventMultiplexer().AddRunsFromDirectory(summaries_dir).Reload()
    data = tf_scalar_data(em)
    plot_tasks(data,
               x_label,
               smoothing_function=smoothing_function,
               max_reward=max_reward,
               xmin=xmin,
               xmax=xmax,
               legend=legend,
               save_directory=save_directory,
               show_plots=show_plots)

def exp_smoothing_weight_test(value):
    """Require that the weight for exponential smoothing is a weight between 0.0 and 1.0"""
    fvalue = float(value)
    if fvalue < 0.0 or fvalue > 1.0:
        raise argparse.ArgumentTypeError("{} must be a float between 0.0 and 1.0".format(value))
    return fvalue

parser = argparse.ArgumentParser()
parser.add_argument("stats_path", metavar="stats", type=str,
                    help="Path to the Tensorflow monitor stats.json file or summaries directory.")
parser.add_argument("--x_label", type=str, default="episode", choices=["episode", "epoch"],
                    help="Whether to use episode or epoch as x label.")
parser.add_argument("--no_legend", action="store_false", default=True, dest="legend",
                    help="Don't show a legend in the plots.")
parser.add_argument("--xmin", type=ge(0), default=None, help="minimum episode for which to show results.")
parser.add_argument("--xmax", type=ge(1), default=None, help="Maximum episode for which to show results.")
parser.add_argument("--max_reward", type=float, default=None, help="Maximum obtainable reward in the environment.")
parser.add_argument("--exp_smoothing", type=exp_smoothing_weight_test, default=None,
                    help="Use exponential smoothing with a weight 0<=w<=1.")
parser.add_argument("--moving_average", type=ge(1), default=None, help="Use a moving average with a window w>0")
parser.add_argument("--save", type=str, dest="save_dir", default=None, help="Save plots to the given directory.")
parser.add_argument("--show", action="store_true", default=False, help="Show plots.")
parser.add_argument(
    '-d', '--debug',
    help="Print debugging statements",
    action="store_const", dest="loglevel", const=logging.DEBUG,
    default=logging.WARNING
)

if __name__ == "__main__":
    args = parser.parse_args()
    logging.basicConfig(level=args.loglevel, format='%(asctime)s - %(levelname)s - %(message)s')
    if args.exp_smoothing is not None and args.moving_average is not None:
        raise WrongArgumentsError("Maximally 1 smoothing technique can be used.")
    smoothing_technique = None
    if args.exp_smoothing is not None:
        smoothing_technique = create_smoother(exponential_smoothing,
                                              args.exp_smoothing)  # args.exp_smoothing holds the weight
    elif args.moving_average is not None:
        smoothing_technique = create_smoother(moving_average,
                                              args.moving_average)  # args.moving_average holds the window
    if os.path.isfile(args.stats_path) and args.stats_path.endswith(".json"):
        plot_gym_monitor_stats(args.stats_path,
                               args.xmax,
                               smoothing_technique,
                               save_directory=args.save_dir,
                               show_plots=args.show)
    elif os.path.isdir(args.stats_path):
        shared_args = {
            "summaries_dir": args.stats_path,
            "xmin": args.xmin,
            "xmax": args.xmax,
            "smoothing_function": smoothing_technique,
            "max_reward": args.max_reward,
            "x_label": args.x_label,
            "legend": args.legend,
            "save_directory": args.save_dir,
            "show_plots": args.show
        }
        plot_tf_scalar_summaries(**shared_args)
    else:
        raise NotImplementedError("Only a Tensorflow monitor stats.json file or summaries directory is allowed.")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="yarll",
    version="0.0.3",
    author="Arno Moonens",
    author_email="arno.moonens@outlook.com",
    description="Yet Another Reinforcement Learning Library",
    license="MIT",
    keywords="deep learning reinforcement learning a3c ddpg sac ppo machine neural networks",
    python_requires=">=3.4",
    install_requires=[
        "numpy",
        "gym>=0.8.0",
        "tensorflow>=1.6.0",
        "matplotlib",
        "scipy",
        "opencv-contrib-python"
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/arnomoonens/YARLL",
    packages=setuptools.find_packages(),
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Topic :: Scientific/Engineering',
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    zip_safe=False
)



# -*- coding: utf8 -*-
import gym
from yarll.environment.wrappers import DescriptionWrapper

class CartPole(DescriptionWrapper):
    """Cartpole-v0 environment wrapper."""

    changeable_parameters: list = [
        {
            "name": "length",
            "type": "range",
            "low": 0.01,
            "high": 5.0
        },
        {
            "name": "masspole",
            "type": "range",
            "low": 0.01,
            "high": 5.0
        },
        {
            "name": "masscart",
            "type": "range",
            "low": 0.01,
            "high": 5.0
        },
    ]

    def __init__(self, length=None, masspole=None, masscart=None, ** kwargs):
        self.length = length
        self.masspole = masspole
        self.masscart = masscart
        super(CartPole, self).__init__(gym.make("OldCartPole-v0"), **kwargs)
        self.change_parameters(length=length, masspole=masspole, masscart=masscart)
        self.metadata["parameters"].update(self.changeable_parameters_values())

    def change_parameters(self, length=None, masspole=None, masscart=None):
        """Change a CartPole environment using a different length and/or masspole."""
        if length is not None:
            self.unwrapped.length = length
        if masspole is not None:
            self.unwrapped.masspole = masspole
        if masscart is not None:
            self.unwrapped.masscart = masscart
        self.unwrapped.total_mass = self.unwrapped.masspole + self.unwrapped.masscart  # Recalculate
        self.unwrapped.polemass_length = self.unwrapped.masspole * self.unwrapped.length  # Recalculate



# -*- coding: utf8 -*-
from typing import Any, List, Optional
import tensorflow as tf
import numpy as np
from yarll.memory.experiences_memory import ExperiencesMemory, Experience
from yarll.misc.utils import RunningMeanStd, normalize


class EnvRunner(object):
    """Environment runner using a policy"""
    def __init__(self, env, policy, config: dict, normalize_states=False, state_preprocessor=None, summary_writer=None) -> None:
        super(EnvRunner, self).__init__()
        self.env = env
        self.policy = policy
        self.state: Optional[np.ndarray] = None
        self.features: Any = policy.initial_features
        self.config: dict = dict(
            batch_update="timesteps",
            episode_max_length=env.spec.max_episode_steps,
            timesteps_per_batch=10000,
            n_iter=100
        )
        self.episode_steps: int = 0
        self.episode_reward: float = 0.0
        self.n_episodes: int = 0
        self.config.update(config)
        self.state_preprocessor = state_preprocessor
        self.summary_writer = summary_writer

        # Normalize states before giving it as input to the network.
        # Mean and std are only updated at the end of `get_steps`.
        self.normalize_states = normalize_states
        if normalize_states:
            self.rms = RunningMeanStd(self.env.observation_space.shape)
        self.reset_env()

    def choose_action(self, state: np.ndarray):
        """Choose an action based on the current state in the environment."""
        return self.policy.choose_action(state, self.features)

    def normalize(self, state: np.ndarray) -> np.ndarray:
        return normalize(state, self.rms.mean, self.rms.std)

    def reset_env(self) -> None:
        """Reset the current environment and get the initial state"""
        self.state = self.env.reset()
        self.state = self.state if self.state_preprocessor is None else self.state_preprocessor(self.state)

    def step_env(self, action):
        """Execute an action in the current environment."""
        state, reward, done, info = self.env.step(self.policy.get_env_action(action))
        state = state if self.state_preprocessor is None else self.state_preprocessor(state)
        return state, reward, done, info

    def get_steps(self, n_steps: int, reset: bool = False, stop_at_trajectory_end: bool = True, render: bool = False) -> ExperiencesMemory:
        if reset:
            self.reset_env()
            self.policy.new_trajectory()
        memory = ExperiencesMemory()
        for _ in range(n_steps):
            input_state = self.normalize(self.state) if self.normalize_states else self.state
            results = self.choose_action(input_state)
            action = results["action"]
            value = results.get("value", None)
            new_features = results.get("features", None)
            new_state, rew, done, _ = self.step_env(action)
            memory.add(self.state, action, rew, value, terminal=done, features=self.features)
            self.state = new_state
            self.features = new_features
            self.episode_reward += rew
            self.episode_steps += 1
            if done or self.episode_steps >= self.config["episode_max_length"]:
                if self.summary_writer is not None:
                    summary = tf.Summary()
                    summary.value.add(tag="global/Episode_length", simple_value=float(self.episode_steps))
                    summary.value.add(tag="global/Reward", simple_value=float(self.episode_reward))
                    self.summary_writer.add_summary(summary, self.n_episodes)
                    self.summary_writer.flush()
                self.episode_reward = 0
                self.episode_steps = 0
                self.n_episodes += 1
                self.reset_env()
                self.features = self.policy.initial_features
                self.policy.new_trajectory()
                # Decide whether to stop when the episode (=trajectory) is done
                # or to keep collecting until n_steps
                if stop_at_trajectory_end:
                    break
            if render:
                self.env.render()

        if self.normalize_states:
            self.rms.add_values([exp.state for exp in memory.experiences])
            for i, exp in enumerate(memory.experiences):
                memory.experiences[i] = Experience(self.normalize(exp.state),
                                                   exp.action,
                                                   exp.reward,
                                                   exp.value,
                                                   exp.features,
                                                   exp.terminal)
        return memory

    def get_trajectory(self, stop_at_trajectory_end: bool = True, render: bool = False) -> ExperiencesMemory:
        return self.get_steps(self.config["episode_max_length"], stop_at_trajectory_end, render)

    def get_trajectories(self, stop_at_trajectory_end: bool = True, render: bool = False) -> List[ExperiencesMemory]:
        """Generate trajectories until a certain number of timesteps or trajectories."""
        use_timesteps = self.config["batch_update"] == "timesteps"
        trajectories = []
        timesteps_total = 0
        i = 0
        while(use_timesteps and timesteps_total < self.config["timesteps_per_batch"]) or\
            (not(use_timesteps) and i < self.config["trajectories_per_batch"]):
            i += 1
            trajectory = self.get_trajectory(stop_at_trajectory_end, render)
            trajectories.append(trajectory)
            timesteps_total += len(trajectory.rewards)
        return trajectories
